/// @file ilmsens_hal.h 
/// @brief Function definitions for Ilmsens HAL API
///
/// This header contains prototypes for the exported functions of the HAL library.
///
/// @author Ralf Herrmann <ralf.herrmann@ilmsens.com>

/* ILMsens UWB HAL API 

This header contains function prototypes for the HAL API.

Platform: C/C++
Author  : R. Herrmann

Revision:

1.0.0: (01/2017)

+ initial revision containing HAL API 
*/

#include <stddef.h> // size_t
#include "ilmsens/ilmsens_error.h"
#include "ilmsens_hal_types.h"
#include "ilmsens_hal_defn.h"

#pragma once

//export and import of API functions
#if defined(WIN32)
#if defined DLL_EXPORT
  #define DECLDIR __declspec(dllexport)
#else
  #define DECLDIR __declspec(dllimport)
#endif
#else
  #define DECLDIR
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////
// UWB HAL API prototypes
///////////////////////////////////////////////////////////////////////////////////////////////////
extern "C"
{
  /**   
   * @defgroup ilmsens_hal_init Library initialization/deinitialization
   * @brief
   * @{
   */  

  /**   
   * @defgroup ilmsens_hal_version Library version and revision
   * @brief
   * @{
   */  

  /** 
   * Return the HAL version.
   *
   * @param pVersion pointer to version information structure
   *
   * @returns ILMSENS_SUCCESS on success 
   * @returns negative @ref ilmsens_error "error-code"
   */
  DECLDIR int ilmsens_hal_getVersion (struct ilmsens_hal_Version *pVersion);

  /** @} ilmsens_hal_version */
 
  /**   
   * @defgroup ilmsens_hal_diag Diagnostic interface
   * @brief
   * @{
   */  
  
  /** 
   * Sets the verbosity of the diagnostics output
   * Diagnostics are sent to std. error output by default.
   * In case of an error, the current level is not changed.
   *
   * @param pLevel verbosity level from @ref ILMSENS_DEB_NO to @ref ILMSENS_DEB_ALL
   *
   * @returns current debug level (on success and error)
   */
  DECLDIR int ilmsens_hal_setDEBLevel (unsigned int pLevel);

  /** @} ilmsens_hal_diag */
  
  /** 
   * Initializes the library.
   * This function must be called before calling any other library function.
   * @returns number of connected devices 
   * @returns negative @ref ilmsens_error "error-code"
   */
  DECLDIR int ilmsens_hal_initHAL   (void);

  /** 
   * Deinitializes the library.
   * Should be called after closing all open devices and before your application terminates.
   * This function always succeeds. Errors that appeared, will only be visible when calling 
   * @ref ilmsens_hal_initHAL() again (i.e. if the library is not unloaded but a new session is started).
   */
  DECLDIR void ilmsens_hal_deinitHAL (void);

  /** @} ilmsens_hal_init */
  
  /**   
   * @defgroup ilmsens_hal_dev Device handling
   * @brief
   * @{
   */  

  /** 
   * Allocates specified devices for a measurement session.
   *
   * This function must be called before doing the configuration or starting a measurement session.
   *
   * @param pDevNums pointer to a first element or an array of device-indexes
   * @param pNum number of array-elements
   * @returns ILMSENS_SUCCESS on success 
   * @returns negative @ref ilmsens_error "error-code"
   */
  DECLDIR int ilmsens_hal_openSensors (unsigned int *pDevNums, unsigned int pNum);

  /** 
   * Releases specified devices.
   *
   * Should be called after the last function call to any of the specified devices.
   *
   * @param pDevNums pointer to a first element or an array of device-indexes
   * @param pNum number of array-elements
   */

  DECLDIR void ilmsens_hal_closeSensors (unsigned int *pDevNums, unsigned int pNum);
  
   /**   
   * @defgroup ilmsens_hal_modinfo Device ID & properties
   * @brief Functions for identification and access to hardware-parameters of connected devices 
   * @{
   */  

  /** 
   * Gets unique device-identifier
   *
   * @param pDevNum device-index
   * @param pID pointer to an output buffer
   * @param pBufferSize Size of buffer pointed to by @p pID in bytes. Maximum ID length is given via @ref ILMSENS_HAL_MOD_ID_BUF_SIZE.
   * @returns number of characters written to the buffer 
   * @returns negative @ref ilmsens_error "error-code"
   */
  DECLDIR int ilmsens_hal_getModId (unsigned int pDevNum, char *pID, size_t pBufferSize);

  /** 
   * Gets device hardware-configuration
   *
   * @param pDevNum device-index
   * @param pInfo pointer to the @ref ilmsens_hal_ModInfo structure
   *
   * @returns ILMSENS_SUCCESS on success (and populates pInfo) 
   * @returns negative @ref ilmsens_error "error-code"
   */
  DECLDIR int ilmsens_hal_getModInfo (unsigned int pDevNum, struct ilmsens_hal_ModInfo *pInfo);

  /** @} ilmsens_hal_modinfo */

  /** @} ilmsens_hal_dev */
  
  /**   
   * @defgroup ilmsens_hal_config Device configuration
   * @brief Functions for configuring specified devices.
   * @{
   */  
  
  /** 
   * Performs the initial setup of specified devices.
   *
   * This function must be called before starting a measurement session.
   *
   * @param pDevNums pointer to a first element or an array of device-indexes
   * @param pNum number of array-elements
   * @param pConfig pointer to configuration parameters in a @ref ilmsens_hal_ModConfig structure
   *
   * @returns ILMSENS_SUCCESS on success
   * @returns negative @ref ilmsens_error "error-code"
   */
  DECLDIR int ilmsens_hal_setupSensors (unsigned int *pDevNums, unsigned int pNum, 
                                        const struct ilmsens_hal_ModConfig *pConfig);

  /** 
   * Sets master/slave operational mode.
   *
   * Application needs to make sure, that any independent sensor
   * is set to master mode.
   *
   * For hardware-synchronized operation of multiple sensors, exactly one sensor 
   * of a connected group has to be master the others in teh group must be configured as slaves.
   *
   * @param pDevNums pointer to a first element or an array of device-indexes
   * @param pNum number of array-elements
   * @param pMode @ref ILMSENS_HAL_SLAVE_SENSOR = slave | @ref ILMSENS_HAL_MASTER_SENSOR = master
   *
   * @returns ILMSENS_SUCCESS on success
   * @returns negative @ref ilmsens_error "error-code"
   */
  DECLDIR int ilmsens_hal_setMaster (unsigned int *pDevNums, unsigned int pNum, int pMode);
  
  /** @} ilmsens_hal_config*/

  /**   
   * @defgroup ilmsens_hal_meas_config Measurement configuration
   * @brief
   * @{
   */  
   
  /** 
   * Sets software averages and wait cycles
   *
   * Software averages define the acquisition aperture duration.
   * If wait cycles are set to 0, devices measure in continuous mode.
   * If wait cycles are non-zero, devices measure in snapshot mode.
   *
   * Note: May only be called when no measurement is running!
   *
   * @param pDevNums pointer to a first element or an array of device-indexes
   * @param pNum number of array-elements
   * @param pAvg number of Sw averages
   * @param pWaitCyc number of wait cycles
   *
   * @returns ILMSENS_SUCCESS on success
   * @returns negative @ref ilmsens_error "error-code"
   */ 
  DECLDIR int ilmsens_hal_setAvg (unsigned int *pDevNums, unsigned int pNum, unsigned int pAvg, 
                                  unsigned int pWaitCyc);
  /** 
   * Resets the M-sequence generator (transmitter) of each device.
   *
   * After digital synchronisation, the M-sequence generator (transmitter) of each device should be reset 
   * to ensure repeatable alignment of the transmitters and receivers. The reset typically takes a few ms to complete 
   * and cannot be used to mute the transmitter(s).
   *
   * Note: May only be called when no measurement is running!
   *
   * @param pDevNums pointer to a first element or an array of device-indexes
   * @param pNum number of array-elements
   *
   * @returns ILMSENS_SUCCESS on success
   * @returns negative @ref ilmsens_error "error-code"
   */                                   
  DECLDIR int ilmsens_hal_setMLBS (unsigned int *pDevNums, unsigned int pNum);
  
  /** 
   * Controls the power-status of the transmitter.
   *
   * Note: May only be called when no measurement is running!
   *
   * @param pDevNums pointer to a first element or an array of device-indexes
   * @param pNum number of array-elements
   * @param pPD @ref ILMSENS_HAL_TX_ON = power-up the transmitter | @ref ILMSENS_HAL_TX_OFF = power-down (mute) the transmitter
   *
   * @returns ILMSENS_SUCCESS on success
   * @returns negative @ref ilmsens_error "error-code"
   */                                   
  DECLDIR int ilmsens_hal_setPD (unsigned int *pDevNums, unsigned int pNum, int pPD);
  
  /** 
   * Performs digital synchronisation. 
   *
   * Must be used at least once before a measurement is started.
   *
   * @param pDevNums pointer to a first element or an array of device-indexes
   * @param pNum number of array-elements
   * @param pMode @ref ILMSENS_HAL_SYNCH_OFF = de-synchronise sensors (revoke) | @ref ILMSENS_HAL_SYNCH_ON = trigger digital synchronisation
   *
   * @returns ILMSENS_SUCCESS on success
   * @returns negative @ref ilmsens_error "error-code"
   */                                   
  DECLDIR int ilmsens_hal_synchMS (unsigned int *pDevNums, unsigned int pNum, int pMode);

  /** @} ilmsens_hal_meas_config */
  
  /**   
   * @defgroup ilmsens_hal_meas_run Measurement run
   * @brief Functions to start/stop measurements with one or more devices.
   *
   * The library supports two modes of operation for transfering measured data from the
   * device-buffer memory to the host-system memory (HAL or application).
   *
   * - In case the measurement session was started in the **raw mode**, measured data is transferred 
   * from the device to the host-system memory whenever a complete measurement is available and
   * the function @ref ilmsens_hal_measRdy() is called by the user application. This way the data transfer
   * is performed synchronously with the user application's thread. However, the user application must
   * ensure fast enough polling of the sensor to prevent ring buffer overflow by itself
   *
   * - In case the measurement session is started in **buffered mode**, a separate thread is created by the
   * library. This thread implements periodic @ref ilmsens_hal_measRdy() "polling" of
   * the devices' ring-buffer status and transfer of measured data to the host". 
   * Therefore, the data transfer is performed asynchronously with respect to the progress of the user 
   * application thread. This way high jitter in processing latency will not cause a discontinuity in 
   * the measurement progress and revent potential overflow of the devices' ring-buffer memory.
   *
   * Note: The **buffered mode** is the preferred mode of operation of the library.
   *
   * @{
   */  

  /** 
   * Starts a measurement run with specified devices.
   *
   * One measurement run may be pending at a time.
   *
   * @param pDevNums pointer to a first element or an array of device-indexes
   * @param pNum number of array-elements
   * @param pMode specifies the mode of operation
   *              (@ref ILMSENS_HAL_RUN_RAW for **raw mode** of operation, 
   *               @ref ILMSENS_HAL_RUN_BUF for **buffered mode** of operation)
   *
   * @returns ILMSENS_SUCCESS on success
   * @returns negative @ref ilmsens_error "error-code"
   *
   * @see @ref ilmsens_hal_measStop()
   */ 
  DECLDIR int ilmsens_hal_measRun (unsigned int *pDevNums, unsigned int pNum, int pMode);
  
  /** 
   * Stops running measurement.
   *
   * @param pDevNums pointer to a first element or an array of device-indexes
   * @param pNum number of array-elements
   *
   * @returns ILMSENS_SUCCESS on success
   * @returns negative @ref ilmsens_error "error-code"
   */ 
    DECLDIR int ilmsens_hal_measStop (unsigned int *pDevNums, unsigned int pNum);

  /** @} ilmsens_hal_meas_run */

  /**   
   * @defgroup ilmsens_hal_meas_data Raw measured data retrieval
   * @brief Functions to read measured data out of internal ring-buffer.
   *
   * - There is a simple blocking interface implemented by @ref ilmsens_hal_measGet().
   *
   * - Furthermode there is a non-blocking interface, implemented by @ref ilmsens_hal_measRdy() 
   *   for polling buffer-level of the ring-buffer(s) and @ref ilmsens_hal_measRead() for actually 
   *   reading the next measurement to the user application's memory.
   *
   * @{
   */  

  /** 
   * Reads the fill-level of the internal ring-buffer for all specified devices.
   *
   * This functions is not blocking and returns immediately. However, in [raw mode](@ref ILMSENS_HAL_RUN_RAW),
   * it will transfer data from a device to the libraries' ring-buffer on the host, if it
   * discovers that a complete dataset is available.
   *
   * @param pDevNums pointer to a first element or an array of device-indexes
   * @param pNum number of array-elements
   *
   * @returns minimum number of complete datasets available for all specified devices
   * @returns 0 if at least one device has no new data available
   * @returns negative @ref ilmsens_error "error-code"
   */  
  DECLDIR int ilmsens_hal_measRdy (unsigned int *pDevNums, unsigned int pNum);

  /** 
   * Reads the measurement data for all specified devices in non-blocking way.
   *
   * This functions is not blocking and returns immediately with the next measurement data 
   * or an error-code if no data are available.
   *
   * @param pDevNums pointer to a first element or an array of device-indexes
   * @param pNum number of array-elements
   * @param pBuffer pointer to output buffer for measured data
   * @param pBufSizeBytes Size of buffer pointed to by @p pBuffer in bytes
   *
   * @returns number of elements (samples) copied to the buffer
   * @returns ILMSENS_ERROR_AGAIN if a completed dataset is not available for every device
   * @returns negative @ref ilmsens_error "error-code"
   *
   * @see @ref ilmsens_hw_data_format
   * @see @ref sec_output_buffer_size
   */  
  DECLDIR int ilmsens_hal_measRead (unsigned int *pDevNums, unsigned int pNum, 
                                    ilmsens_hal_SampleType *pBuffer, size_t pBufSizeBytes);

  /** 
   * Blocks and reads the measurement data for all specified devices when it becomes available.
   *
   * This functions blocks the caller until at least one complete measurement is available
   * for every device or a specified timeout expired.
   * The buffer @p pBuffer must be large enough to hold one complete dataset for each device, i.e.
   * it must be able to hold at least @p pNum complete datasets.
   *
   * Note: if @p pTimeoutMillis is 0, this function will block forever...
   *
   * @param pDevNums pointer to a first element or an array of device-indexes
   * @param pNum number of array-elements
   * @param pBuffer pointer to output buffer for measured data
   * @param pBufSizeBytes Size of buffer pointed to by @p pBuffer in bytes
   * @param pTimeoutMillis timeout [ms] (0 = forever)
   *
   * @returns number of elements (samples) copied to the buffer
   * @returns ILMSENS_ERROR_TIMEOUT when the timeout expired before enough data was received
   * @returns negative @ref ilmsens_error "error-code"
   *
   * @see @ref ilmsens_hw_data_format
   * @see @ref sec_output_buffer_size
   */  
  DECLDIR int ilmsens_hal_measGet (unsigned int *pDevNums, unsigned int pNum, 
                                   ilmsens_hal_SampleType *pBuffer, size_t pBufSizeBytes, 
                                   unsigned int pTimeoutMillis);

  /** @} ilmsens_hal_meas_data */
  
  
  /**   
   * @defgroup ilmsens_hal_lowlevel Low-level access to sensor registers & memory
   * @brief
   * @{
   */  

  /** 
   * Reads value from register at @p pReg address from all specified devices to buffer.
   * The buffer @p pVal must be large enough to hold one word for each device, i.e.
   * it must be at least @p pNum words in size.
   *
   * @param pDevNums pointer to a first element or an array of device-indexes
   * @param pNum number of array-elements
   * @param pReg register address
   * @param pVal pointer to buffer for register value(s)
   * @param pBufSizeBytes Size of buffer pointed to by @p pVal in bytes
   *
   * @returns number of words (elements) copied to the buffer
   * @returns negative @ref ilmsens_error "error-code"
   */ 
  DECLDIR int ilmsens_hal_readReg (unsigned int *pDevNums, unsigned int pNum, unsigned int pReg, 
                                   ilmsens_hal_MemoryType *pVal, size_t pBufSizeBytes);

  /** 
   * Writes @p pVal value to register at @p pReg address to all specified devices.
   * The same value is written to each device.
   *
   * @param pDevNums pointer to a first element or an array of device-indexes
   * @param pNum number of array-elements
   * @param pReg register address
   * @param pVal new register value
   *
   * @returns ILMSENS_SUCCESS on success
   * @returns negative @ref ilmsens_error "error-code"
   */ 
  DECLDIR int ilmsens_hal_writeReg (unsigned int *pDevNums, unsigned int pNum, unsigned int pReg, 
                                    ilmsens_hal_MemoryType pVal);

  /** 
   * Reads @p pNumEl elements (32-bit words) starting at address @p pAdr from the internal memory 
   * of specified devices into a buffer @p pVal. The buffer must be large enough to hold @p pNumEl
   * words for all specified devices, i.e. it provide space for at least @p pNumEl x @p pNum words.
   *
   * @param pDevNums pointer to a first element or an array of device-indexes
   * @param pNum number of array-elements
   * @param pAdr word-aligned start memory address
   * @param pNumEl number of words (elements) to read
   * @param pVal pointer to buffer for the transferred memory content
   * @param pBufSizeBytes Size of buffer pointed to by @p pVal in bytes
   *
   * @returns number of words (elements) copied to the buffer
   * @returns negative @ref ilmsens_error "error-code"
   */ 
  DECLDIR int ilmsens_hal_readBlk (unsigned int *pDevNums, unsigned int pNum, 
                                   unsigned int pAdr, unsigned int pNumEl, 
                                   ilmsens_hal_MemoryType *pVal, size_t pBufSizeBytes);

  /** 
   * Writes @p pNumEl elements (32-bit words) from buffer @p pVal to internal memory starting at address @p pAdr
   * of specified devices. The same content is written to each device, i.e. the buffer @p pVal must hold @p pNumEl words
   * regardless of @p pNum.
   *
   * @param pDevNums pointer to a first element or an array of device-indexes
   * @param pNum number of array-elements
   * @param pAdr word-aligned start memory address
   * @param pNumEl number of words (elements) to write
   * @param pVal pointer to buffer with data to write
   * @param pBufSizeBytes Size of buffer pointed to by @p pVal in bytes
   *
   * @returns ILMSENS_SUCCESS on success
   * @returns negative @ref ilmsens_error "error-code"
   */ 
  DECLDIR int ilmsens_hal_writeBlk (unsigned int *pDevNums, unsigned int pNum, 
                                    unsigned int pAdr, unsigned int pNumEl, 
                                    ilmsens_hal_MemoryType *pVal, size_t pBufSizeBytes);
  
  /** @} ilmsens_hal_lowlevel */

}
